// Copyright (c)1999-2001 by "GMD-AiS"
//  Full name: "GMD National Research Center for Information Technology -
//  Institute for Autonomous Intelligent Systems", in future
//  named "Fraunhofer Institute for Autonomous Intelligent Systems"
//
// This software is provided 'as is' by the above copyright holder without
// any explicit or implicit warranty and support.
//
// Within the granted license, the licensee is allowed to freely use
// and modify the software EXCLUSIVELY FOR academical, SCIENTIFIC, and/or
// educational PURPOSES. It is permitted to cite (in the common scientific
// sense) fragments of the code in scientific publications of the licensee,
// and also use modified or unmodified parts of the software in own
// research prototypes.
//
// Commerical use, as well as integration into hardware/software systems or
// other products, other then for aforementioned scientific purposes, is
// PROHIBITED WITHOUT EXPLICIT PERMISSION of the copyright holder.
//
// Prototyping software that makes use of parts of this code have to
// refer to the above copyright when given out to third parties by
// a line
//   "This software contains or uses code that is copyrighted
//    by the GMD National Research Center for Information Technology
//    (Fraunhofer Institute for Autonomous Intelligent Systems)"
// in the user visible part of the software (documentation, about box, etc.)
//
// The copyright holder provides the software 'as is' without any support
// (request for bugfixes, for instance), and without any implicit or explicit
// warranty. In particular, the copyright holder cannot be made liable
// for lack of certain features, for damages of any kind that might result
// from any proper or improper use of this software. This software is not
// to be used in safety-critical environments.
//
// Beyond this license, GMD-AiS appreciates, in the scientific sense,
// reports and code submissions of significant changes or improvements
// of the jFTA core software that might be integrated into a future
// version of this code. Such reports are entirely voluntary, though;
// they are not mandated by this license. Use the address jfta@borneo.gmd.de
// (in future jfta@borneo.ais.fhg.de) for submissions.

// $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $

//
// Generic system configurator, reads a HOST/TB/EN/AC/TAG/SYNC configuration
// from a file
//
package jfta.actors;

import jfta.exception.*;
import jfta.tags.*;

import java.io.*;
import java.util.*;

/**
 * Class <code>FTASysConfigInit</code> is a default init actor that
 * reads its configuration from a file.
 * @author   Holger Veit
 * @version  $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
 * @see      jfta.actors.FTAInitActor
 * @see      jfta.actors.FTAActorI
 * @since    JFTA1.0
 */
public class FTASysConfigInit extends FTAInitActor implements FTAActorI
{
    /**
     * internally used
     */
    private final static int ENTRY = 0;
    private final static int HOST_ENTRY = 1;
    private final static int TB_ENTRY = 2;
    private final static int EN_ENTRY = 3;
    private final static int AC_ENTRY = 4;
    private final static int TG_ENTRY = 5;
    private final static int ET_ENTRY = 6;
  
    /**
     * @serial stores the parsed entries
     */
    private Vector entries;
    /**
     * @serial internally used to store accepted file commands
     */
    private Hashtable commands;
    /**
     * @serial internally used 
     */
    private boolean verbose;

    /**
     * generates a standardized exception during parsing.
     * Not for general use.
     * @param st the stream tokenizer used for parsing a line
     * @param msg the exception message to be printed
     * @exception FTAArgumentException the generated exception
     */
    private final void exception(StreamTokenizer st,String msg)
	throws FTAArgumentException {
	userMessage(LOG_ERROR,msg);
	throw new FTAArgumentException("Line " + 
				       (st==null?0:st.lineno()) +
				       ": " + msg);
    }

    /**
     * generates a standardized exception during parsing.
     * Not for general use.
     * @param st the stream tokenizer used for parsing a line
     * @param e exception raised
     * @exception FTAArgumentException the generated exception
     */
    private final void exception(StreamTokenizer st,Exception e)
	throws FTAArgumentException {
	exception(st,e.getMessage());
    }


    /**
     * Generates a standardized exception during execution of
     * the parsed command list. Not for general use.
     * @param st the stream tokenizer used for parsing a line
     * @exception FTAEvalActorException the generated exception
     */
    private final void evalexception(String msg)
	throws FTAEvalActorException {
	userMessage(LOG_ERROR,msg);
	throw new FTAEvalActorException("FTASysConfigInit: "+msg);
    }

    /**
     * Output a message when verbose parsing is enabled.
     * @param msg the message to be printed
     */
    private final void verbosemsg(String msg) {
	if (verbose)
	    userMessage(LOG_STD,msg);
    }

    /**
     * Output a message when verbose parsing is enabled.
     * @param msg the message to be printed
     */
    private final void verbosemsg_suc(String msg) {
	if (verbose)
	    userMessage(LOG_SUCCESS,msg);
    }

    /**
     * Internally used base class to express different parsable entities
     * Effectively the <i>Command</i> design pattern
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     */
    class Entry extends Object {
	public Entry() {
	    super();
	}
    
	/**
	 * Parse a String
	 * @param st the StreamTokenizer that returns the next token. This is
	 *   expected to be any String. Strings that contain whitespace are
	 *   expected to be enclosed in double quotes <code>""</code>
	 * @return the parsed String
	 * @exception FTAArgumentException thrown when there is an unexpected
	 *   end of line, or the parsed argument is not a valid string
	 */
	protected String get_string(StreamTokenizer st)
	    throws FTAArgumentException {
	    try {
		if (st.nextToken() != StreamTokenizer.TT_WORD &&
		    st.nextToken() != '"')
		    exception(st,"Expected a string");
	    } catch (IOException e) {
		exception(st,e);
	    }
	    return _svc.expandEnv(st.sval);
	}

	/**
	 * Parse an integer
	 * @param st the StreamTokenizer that returns the next token. This is
	 *   expected to be a numeric integer value
	 * @return the parsed integer
	 * @exception FTAArgumentException thrown when there is an unexpected
	 *   end of line, or the parsed argument is not an integer
	 */
	protected int get_int(StreamTokenizer st) throws FTAArgumentException {
	    try {
		String str = get_string(st);
		return Integer.parseInt(str);
	    } 
	    catch (NumberFormatException e) {
		exception(st,"Expected a number");
	    }
	    return 0;
	}
    

	/**
	 * Parse a boolean.
	 * @param st the StreamTokenizer that returns the next token. This is
	 *   expected to be either 'true' or 'false'
	 * @return the parsed boolean
	 * @exception FTAArgumentException thrown when there is an unexpected
	 *   end of line, or the parsed argument is not 'true' or 'false'
	 */
	protected boolean get_boolean(StreamTokenizer st)
	    throws FTAArgumentException {
	    String str = get_string(st);
	    if (str.equals("false"))
		return false;
	    else if (str.equals("true"))
		return true;
	    else
		exception(st,"Expected true/false");
	    return false;
	}
    
	/**
	 * anachronism: parse an integer which can be also true=1 or false=0
	 * @param st the StreamTokenizer that returns the next token. This is
	 *   expected to be either 'true' or 'false'
	 * @return the parsed boolean
	 * @exception FTAArgumentException thrown when there is an unexpected
	 *   end of line, or the parsed argument is not 'true' or 'false'
	 */
	protected int get_boolint(StreamTokenizer st)
	    throws FTAArgumentException {
	    try {
		String str = get_string(st);
		if (str.equals("false"))
		    return 0;
		else if (str.equals("true"))
		    return 1;
		else
		    return Integer.parseInt(str);
	    } catch (NumberFormatException e) {
		//e.printStackTrace();
		exception(st,e);
	    }
	    return 0;
	}

	/**
	 * Create the object specified by this entry. This is called by
	 * <code>FTASysConfigInit</code> when all entries have been parsed.
	 * @param ac a reference to the FTASysConfigInit class to execute
	 *   the requested commands
	 * @exception FTAEvalActorException thrown when the operation to be
	 *   performed goes wrong.
	 * @see jfta.actors.FTASysConfigInit FTASysConfigInit class
	 */
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    evalexception("Software error: baseclass called");
	}
    
	/**
	 * Parse the remaining arguments of the command line (command token
	 * already consumed) and create a new specific Entry object that
	 * holds the complete information.
	 * @return a new Entry for this command type.
	 * @param st StreamTokenizer that returns the arguments as tokens
	 * @exception FTAArgumentException thrown in case of an incomplete or
	 *   improper argument in context.
	 */
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    exception(null,"Software error: baseclass called");
	    return null;
	}
    
	/**
	 * Function to be implemented for a new command.
	 * @return integer constant for the type of entry. This constant
	 * is supposedly declared as a public static constant in FTASysConfigInit.
	 */
	public int isA() {
	    return ENTRY;
	}
    }
  
    /**
     * Internally used class to parse, store and create a TAGBOARD
     * Uses the <i>Command</i> design pattern.
     * Syntax of the parsed line:
     * <code>TAGBOARD name host true/false</code>
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     * @see #Entry class Entry
     */
    class TbEntry extends Entry {
	private String name;
	private String host;
	private int dist;
    
	public TbEntry() {
	    super();
	}
    
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    TbEntry tbe = new TbEntry();
	    tbe.name = get_string(st);
	    tbe.host = get_string(st);
	    tbe.dist = get_boolint(st);
	    return tbe;
	}
    
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    try {
		addTagboard(name,host,dist);
		verbosemsg("Create TB " + name +
			   " on host " + host +
			   " (" + ((dist&1)==1?"":"non") +
			   "distributed)");
	    }
	    catch (FTAException e) {
		evalexception("Failed to create TB "+name);
	    }
	}
    
	public int isA() {
	    return TB_ENTRY;
	}
    }
  
    /**
     * Internally used class to parse, store and create a HOST
     * Uses the <i>Command</i> design pattern.
     * Syntax of the parsed line:
     * <code>HOST host</code>
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     * @see #Entry class Entry
     */
    class HoEntry extends Entry {
	private String host;
    
	public HoEntry() {
	    super();
	}
    
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    HoEntry hoe = new HoEntry();
	    hoe.host = get_string(st);
	    return hoe;
	}
    
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    try {
		addHost(host);
		verbosemsg("Create HOST " + host);
	    }
	    catch (FTAException e) {
		evalexception("Failed to create HOST "+host);
	    }
	}
    
	public int isA() {
	    return HOST_ENTRY;
	}
    }
  
    /**
     * Internally used class to parse, store and create an ENSEMBLE
     * Uses the <i>Command</i> design pattern.
     * Syntax of the parsed line:
     * <code>ENSEMBLE name host true/false</code>
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     * @see #Entry class Entry
     */
    class EnEntry extends Entry {
	private String name;
	private String host;
	private int dist;
    
	public EnEntry() {
	    super();
	}
    
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    EnEntry ene = new EnEntry();
	    ene.name = get_string(st);
	    ene.host = get_string(st);
	    ene.dist = get_boolint(st);
	    return ene;
	}
    
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    try {
		addEnsemble(name,host,dist);
		verbosemsg("Create EN " + name +
			   " on host " + host +
			   " (" + ((dist&1)==1?"":"non") +
			   "distributed)");
	    }
	    catch (FTAException e) {
		evalexception("Failed to create EN "+name);
	    }
	}
    
	public int isA() {
	    return EN_ENTRY;
	}
    }
  
    /**
     * Internally used class to parse, store and create an ACTOR
     * Uses the <i>Command</i> design pattern.
     * Syntax of the parsed line:
     * <code>ACTOR host enname acname actype arglist[]</code>
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     * @see #Entry class Entry
     */
    class AcEntry extends Entry {
	private String host;
	private String enname;
	private String acname;
	private String actype;
	private String[] args;
    
	public AcEntry() {
	    super();
	}
    
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    AcEntry ace = new AcEntry();
	    ace.host = get_string(st);
	    ace.enname = get_string(st);
	    ace.acname = get_string(st);
	    ace.actype = get_string(st);

	    int i = 0, tok;
	    Vector argv = new Vector();
	    try {
		while ((tok = st.nextToken()) != StreamTokenizer.TT_EOF) {
		    switch (tok) {
		    case StreamTokenizer.TT_WORD:
		    case '\"':
			argv.addElement(new String(st.sval));
			break;
		    case StreamTokenizer.TT_NUMBER:
			argv.addElement(new String(Double.toString(st.nval)));
			break;
		    case StreamTokenizer.TT_EOL:
			ace.args = new String[i];
			argv.copyInto(ace.args);
			st.pushBack();
			return ace;
		    default:
			exception(st,"Expected EOL");
		    }
		    i++;
		}
	    }
	    catch (IOException e) {
		exception(st,e);
	    }
	    return ace;
	}
    
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    try {
		addActor(host,enname,acname,actype,args);
		verbosemsg("Add ACTOR " + acname +
			   " (type " + actype +
			   ") on host " + host +
			   " to EN " + enname);
	    }
	    catch (FTAException e) {
		evalexception("Failed to create ACTOR " + 
					       acname +
					       ", error="+e);
	    }
	}
    
	public int isA() {
	    return AC_ENTRY;
	}
    }
  
    /**
     * Internally used class to parse, store and create a TAG
     * Uses the <i>Command</i> design pattern.
     * Syntax of the parsed line:
     * <code>TAG tbname tgtype arglist[]</code>
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     * @see #Entry class Entry
     */
    class TgEntry extends Entry {
	private String tbname;
	private String tgtype;
	private String[] args;
    
	public TgEntry() {
	    super();
	}
    
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    TgEntry tge = new TgEntry();
	    tge.tbname = get_string(st);
	    tge.tgtype = get_string(st);
      
	    int i = 0, tok;
	    Vector argv = new Vector();
	    try {
		while ((tok = st.nextToken()) != StreamTokenizer.TT_EOF) {
		    switch (tok) {
		    case StreamTokenizer.TT_WORD:
		    case '\"':
			argv.addElement(new String(st.sval));
			break;
		    case StreamTokenizer.TT_NUMBER:
			argv.addElement(new String(Double.toString(st.nval)));
			break;
		    case StreamTokenizer.TT_EOL:
			tge.args = new String[i];
			argv.copyInto(tge.args);
			st.pushBack();
			return tge;
		    default:
			exception(st,"Expected EOL");
		    }
		    i++;
		}
	    }
	    catch (IOException e) {
		//e.printStackTrace();
		exception(st,e);
	    }
	    return tge;
	}
    
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    try {
		putTag(tbname,tgtype,args);
		verbosemsg("Put TAG (type " + tgtype +
			   ") to TB " + tbname);
	    }
	    catch (FTAException e) {
		evalexception("Failed to create TAG at TB "+tbname);
	    }
	}
    
	public int isA() {
	    return TG_ENTRY;
	}
    }
  
    /**
     * Internally used class to parse, store and create an ETSYNC
     * Uses the <i>Command</i> design pattern.
     * Syntax of the parsed line:
     * <code>ETSYNC enname tbname</code>
     * @author Holger Veit
     * @version $Id: FTASysConfigInit.java,v 1.5 2001/06/21 14:57:09 veit Exp $
     * @see #Entry class Entry
     */
    class EtEntry extends Entry {
	private String enname;
	private String tbname;
    
	public EtEntry() {
	    super();
	}
    
	public Entry read_args(StreamTokenizer st) throws FTAArgumentException {
	    EtEntry ete = new EtEntry();
	    ete.enname = get_string(st);
	    ete.tbname = get_string(st);
	    return ete;
	}
    
	public void eval(FTASysConfigInit ac) throws FTAEvalActorException {
	    try {
		addETSync(enname,tbname);
		verbosemsg("Add ETSYNC between EN " + enname +
			   " and TB " + tbname);
	    }
	    catch (FTAException e) {
		evalexception("Failed to create ETSYNC for " +
			      tbname + "/" + enname);
	    }
	}
    
	public int isA() {
	    return ET_ENTRY;
	}
    }
  
    /**
     * Read a configuration file and store the parsed lines as
     * <code>Entry</code> objects in the <code>entries<</code> vector.
     * @param cfgfile The pathname of the configuration file
     * @exception FTAArgumentException thrown when there is an error in
     *   one of the command lines.
     * @see #Entry class Entry
     */
    private void read_file(String cfgfile) throws FTAArgumentException {
	int lineno = 0;
	try {
	    //System.out.println("cfgfile="+cfgfile);
	    FileReader fr = new FileReader(cfgfile);
	    StreamTokenizer st = new StreamTokenizer(fr);
	    st.resetSyntax();
	    st.eolIsSignificant(true);
	    //st.parseNumbers();
	    st.commentChar('#');
	    st.slashStarComments(true);
	    st.slashSlashComments(true);
	    st.whitespaceChars(0,' ');
	    st.quoteChar('"');
	    st.wordChars('!','!');
	    st.wordChars('$','~');
      
	    int tok;
      
	    while ((tok = st.nextToken()) != StreamTokenizer.TT_EOF) {
		lineno = st.lineno();
		if (tok == StreamTokenizer.TT_WORD) {
		    String keywd = st.sval.toLowerCase();
		    Entry e = (Entry)commands.get(keywd);
		    if (e == null) {
			exception(st,"Invalid token: "+keywd);
		    } else {
			Entry obj = e.read_args(st);
			entries.add(obj);
		    }
		} else if (tok== StreamTokenizer.TT_EOL) {
		    continue;
		} else {
		    exception(st,"Invalid Keyword: not a token " + tok);
		}
		// expect at this point a parsed line
		tok = st.nextToken();
		if (tok == StreamTokenizer.TT_EOF)
		    break;
		else if (tok == StreamTokenizer.TT_EOL)
		    continue;
		else
		    exception(st,"Expected EOL");
	    }
	    fr.close();
	}
	catch(IOException e) {
	    //e.printStackTrace();
	    exception(null,e);
	}
    }

    /**
     * Iterates through all stored entries and calls the <code>eval()</code>
     * method of all <code>Entry</code> object that match the specified
     * objtype.
     * @param objtype The object type to match.
     * @exception FTAEvalActorException thrown if the eval method of an
     *   Entry fails.
     * @see #Entry class Entry
     */
    private void setup_object(int objtype) throws FTAEvalActorException {
	Enumeration iter = entries.elements();
	while (iter.hasMoreElements()) {
	    Entry e = (Entry)iter.nextElement();
	    if (e.isA() == objtype) {
		e.eval(this);
	    }
	}
    }
  
    /**
     * This is the central routine of an actor, which will actually create
     *   the parsed objects.
     * @exception FTAEvalActorException thrown when one of the creation
     *   operations fails.
     * @return the returncode
     *   <dt>TERMINATE<dd>return code to terminate this actor
     *   after this tick.</dl>
     */
    public int eval() throws FTAEvalActorException {
	verbosemsg_suc("Setting up FTA system");
	setup_object(HOST_ENTRY);
	setup_object(TB_ENTRY);
	setup_object(EN_ENTRY);
	setup_object(AC_ENTRY);
	setup_object(TG_ENTRY);
	setup_object(ET_ENTRY);
	verbosemsg_suc("FTA system completely setup");
	return FTAActor.TERMINATE;
    }

    /**
     * The public constructor of the class.
     * This will initialize the internal fields and register a number
     * of understood commands.
     */
    public FTASysConfigInit() {
	super();
    
	entries = new Vector();
    
	commands = new Hashtable();
	commands.put("host",new HoEntry());
	commands.put("tagboard",new TbEntry());
	commands.put("ensemble",new EnEntry());
	commands.put("actor",new AcEntry());
	commands.put("tag",new TgEntry());
	commands.put("etsync",new EtEntry());
	verbose = false;
    }
  
    /**
     * This is the initialization routine of the actor, which will accept
     * a filename and an optional <code>--verbose</code> argument.
     * @param args List of arguments.
     *   <ul><li>args[0] is the pathname of the configuration file.
     *   <li>args[1] (optional) is the String <code>--verbose</code>
     * @exception FTAArgumentException thrown when the arguments ar missing
     *   or invalid.
     */
    public void init(String[] args) throws FTAArgumentException {
	if (args.length < 1)
	    exception(null,"FTASysConfigInit: Expected Filename as argument");
	if (args.length == 2 && args[1].equals("--verbose"))
	    verbose = true;

	read_file(args[0]);
    }
}

/*
 * Local variables: 
 * tab-width: 8
 * c-basic-offset: 4
 * End: 
 */
