// Copyright (c)1999-2001 by "GMD-AiS"
//  Full name: "GMD National Research Center for Information Technology -
//  Institute for Autonomous Intelligent Systems", in future
//  named "Fraunhofer Institute for Autonomous Intelligent Systems"
//
// This software is provided 'as is' by the above copyright holder without
// any explicit or implicit warranty and support.
//
// Within the granted license, the licensee is allowed to freely use
// and modify the software EXCLUSIVELY FOR academical, SCIENTIFIC, and/or
// educational PURPOSES. It is permitted to cite (in the common scientific
// sense) fragments of the code in scientific publications of the licensee,
// and also use modified or unmodified parts of the software in own
// research prototypes.
//
// Commerical use, as well as integration into hardware/software systems or
// other products, other then for aforementioned scientific purposes, is
// PROHIBITED WITHOUT EXPLICIT PERMISSION of the copyright holder.
//
// Prototyping software that makes use of parts of this code have to
// refer to the above copyright when given out to third parties by
// a line
//   "This software contains or uses code that is copyrighted
//    by the GMD National Research Center for Information Technology
//    (Fraunhofer Institute for Autonomous Intelligent Systems)"
// in the user visible part of the software (documentation, about box, etc.)
//
// The copyright holder provides the software 'as is' without any support
// (request for bugfixes, for instance), and without any implicit or explicit
// warranty. In particular, the copyright holder cannot be made liable
// for lack of certain features, for damages of any kind that might result
// from any proper or improper use of this software. This software is not
// to be used in safety-critical environments.
//
// Beyond this license, GMD-AiS appreciates, in the scientific sense,
// reports and code submissions of significant changes or improvements
// of the jFTA core software that might be integrated into a future
// version of this code. Such reports are entirely voluntary, though;
// they are not mandated by this license. Use the address jfta@borneo.gmd.de
// (in future jfta@borneo.ais.fhg.de) for submissions.

//Source file: $Id: FTADaemon.java,v 1.14 2001/06/21 14:57:10 veit Exp $

package jfta.bootstrap;

import java.net.*;
import java.util.*;
import java.io.*;

import jfta.base.*;
import jfta.http.*;
import jfta.system.*;
import jfta.exception.*;

/**
 * Daemon is the remote process that is called to instantiate
 * a new FTA subsystem on that host.
 * It can create the first SysSub, as well as secondary ones
 */
public class FTADaemon {
    
    private final static String VERSION = "1.1";

    public final static int LOG_STD = FTAServerLogServlet.LOG_STD;
    public final static int LOG_ERROR = FTAServerLogServlet.LOG_ERROR;
    public final static int LOG_WARNING = FTAServerLogServlet.LOG_WARNING;
    public final static int LOG_SUCCESS = FTAServerLogServlet.LOG_SUCCESS;
    public final static int LOG_USER = FTAServerLogServlet.LOG_USER;

    /**
     * local service instance
     */
    private static FTAService _svc = FTAService.Instance();

    /**
     * the multicast address for FTADaemon messages
     */
    private FTABroadcastChannel _ftadcmds;

    /**
     * message listener for FTADaemon communication
     */
    private FTAListenerAdapter _ftadlistener;

    /**
     * the daemon port
     */
    private static int _ftadport;

    /**
     * my hostname
     */
    private String _myhostname;

    /**
     * my HTTP server
     */
    private FTAHttpServer _myserver;

    /**
     * my threadgroup
     */
    private ThreadGroup _mytg;

    /**
     * the list of local offices
     */
    private Hashtable _offices;
  
    /**
     * the list of remote systems
     */
    private Hashtable _remotesystems;
  
    /**
     *  the list of other daemons
     */
    private Hashtable _daemons;

    /**
     * instance pointer (singleton pattern)
     */
    private static FTADaemon _instance = null;
  
    /**
     * my URL
     */
    private String _myurl;
  
    /**
     * is a dirserver?
     */
    private boolean _isdirserver = false;

    /**
     * the URL of the directory server
     */
    private String _dirserver;

    /**
     * directory properties
     */
    private Properties _dir;

    /**
     * directory file
     */
    File _dirfile;

    /**
     * address pool for multicast
     */
    private String _addrpool;

    /**
     * communication socket for directory server
     */
    private ServerSocket _ss;

    /**
     * # of creations in progress
     */
    private FTASemaphore _createsema;

    /**
     * table of init actors
     */
    private FTAPropertyFile _initactors;

    /**
     * update latency constant
     */
    private int _latency;

    /**
     * @roseuid 383EAE42FAD4
     */
    public static FTADaemon Instance() {
	if (_instance == null)
	    _instance = new FTADaemon();
	return _instance;
    }
  
    /**
     * This is the main program of the distributed server.
     * It will create an FTADaemon object.
     * @roseuid 383EAE436931
     */
    public static void main(String[] args) {
    
	FTADaemon myself = null;
    
	// create a new instance of the daemon
	myself = FTADaemon.Instance();
    
	System.out.println("FTADaemon: running on " + 
			   myself._myhostname);
    }
  
    /**
     * @roseuid 383EAE452DE8
     */
    protected FTADaemon() {
	super();

	// register myself in this FTAService instance
	_svc.setDaemon(this);

	// set hostname
	_myhostname = _svc.getFQDNname();
    
	// initialize data structures
	_offices = new Hashtable();
	_remotesystems = new Hashtable();
	_daemons = new Hashtable();
	_createsema = new FTASemaphore();
	_initactors = 
	    new FTAPropertyFile(System.getProperty("user.home") +
				File.separator + "jfta.initactors");

	// my thread group
	_mytg = new ThreadGroup("FTADaemon");

	// the daemon port
	_ftadport = _svc.FTADAEMONPORT;

	// some params
	_latency =
	    _svc.getIntegerProperty("fta_daemon_update_latency",5);

	// my URL
	_myurl = buildURL(_myhostname,"");

	// retrieve the multicast address pool
	_addrpool = _svc.FTADAEMONADDRPOOL;

	// set the own FTADaemon Multicast
	InetAddress addr = _svc.FTADAEMONMULTIADDR;
	int ttl = _svc.FTADAEMONTTL;
	
	// setup an own message listener
	_ftadcmds = 
	    _svc.getMulticastChannel(addr,_ftadport,ttl,
				     FTAMsg.CMDSIZE);

	// add a msghandler
	_ftadlistener = new FTADaemonListener();

	// load HTTP server config
	String cfgfile = _svc.expandEnv("${jftaroot}/jfta/bootstrap/httpd.cfg");
	Properties httpdcfg = new Properties();
	try {
	    // load properties
	    httpdcfg.load(new FileInputStream(cfgfile));

	    // start HTTP server
	    _myserver = new FTAHttpServer(this,httpdcfg,_mytg);
	}
	catch (IOException e) {
	    e.printStackTrace();
	    System.out.println("Http server config file "+cfgfile+
			       " missing or in error");
	    System.exit(1);
	}
	catch (Exception e) {
	    System.out.println("Http server could not be started");
	    System.exit(1);
	}

	// become a directory server
	String dirhost = _svc.getProperty("fta_directory_server_host",
					  null);
	if (dirhost != null && dirhost.equals(_myhostname)) {
	    _isdirserver = true;

	    // obtain Dirserver properties
	    try {
		_dirserver = buildURL(dirhost,"");

		// get dirserver properties
		_dirfile = new File(System.getProperty("user.home") +
				    File.separator + "jfta.directory");
		_dir = new Properties();
		_dir.load(new FileInputStream(_dirfile));
	    }
	    catch(IOException e) {
		System.out.println("Failed to load directory, creating new one");
		try {
		    save();
		}
		catch(IOException e2) {
		    System.out.println("Terminal failure to create directory");
		    System.exit(1);
		}
	    }

	    System.out.println("Becoming the directory server");
	}

	/* finally, update the own tables */
	long now = System.currentTimeMillis();
	_ftadcmds.sendMsg(FTADaemonReqMsg.get(now));
	_ftadcmds.sendMsg(FTADirReqMsg.get(now));
    }
  
    /*package*/ FTAPropertyFile getInitActors() {
	return _initactors;
    }

    /********************************************************************/
    /* DirServer support */
    /********************************************************************/

    private void save() throws IOException {
	_dir.save(new PrintStream(new FileOutputStream(_dirfile)),
		  "Automatically generated - Do not modify!");
    }

    /**package*/ void resetAddrs() throws IOException {
	_dir.clear();
	save();
    }

    /*package*/ String getNewAddr(String from) {
	for (int n1=0; n1<255; n1++) {
	    for (int n2=1; n2<254; n2 += 2) {
		String s = _addrpool + n1 + "." + n2;
		if (_dir.get(s) == null) {
		    Date d = new Date();
		    _dir.put(s,from+"/"+d);
		    try {
			save();
			return s;
		    }
		    catch(IOException e) {
			return null;
		    }
		}
	    }
	}
	return null;
    }

    /*package*/ String getAddrPool() {
	return _addrpool;
    }

    /*package*/ int getAddrPoolSize() {
	return _dir.size();
    }

    /*package*/ boolean removeAddr(String addr) throws IOException {
	// invalid parameter?
	if (addr == null)
	    return false;

	// discard it
	_dir.remove(addr);

	// save properties
	save();
	return true;
    }

    /********************************************************************/
    /* FTA Daemon functions */
    /********************************************************************/

    /**
     * Create a new FTA system with the name 'sysname'
     * When created, it will run the standalone actor acname once
     * @roseuid 383EAE42FAD5
     */
    /*package*/ boolean createSystem(String sysname, String acname, 
				      String[] arglist) {

	_myserver.logMessage(LOG_STD,
			     sysname + 
			     ": create new system -- may need some time");

	// unblock creation sema (warning: this is a bug in general:
	// if some other creation is in progress, it will be unblocked
	// in controlsystem. Ignoring this for now)
	_createsema.reset();
	// already existing?
	if (_offices.containsKey(sysname)) {
	    _myserver.logMessage(LOG_ERROR,
				 sysname+
				 ": create System failed (exists)");
	    return false;
	}
    
	Thread th = new SysThread(this,sysname,_myhostname,
				  acname,arglist,0,0);
	th.start();

	//System.out.println("FTADaemon: createSystem="+sysname+" done");
	return true;
    }
  
    /**
     * create a secondary syssub (new host)
     */
    public boolean addSysSub(String sysname,String mainhost,
			     int backport,int hostnum) {
	// already existing?
	//System.out.println("FTADaemon: addSysSub="+mainhost+'@'+sysname);
	try {
	    FTASysSub ss = (FTASysSub)_offices.get(sysname);
	    if (ss != null) {
		_myserver.logMessage(LOG_ERROR,
				     sysname+
				     ": create SysOffice failed (exists)");
		return false;
	    }

	    Thread th = new SysThread(this,sysname,mainhost,
				      null,null,
				      backport,hostnum);
	    th.start();

	    _myserver.logMessage(LOG_SUCCESS,
				 sysname+
				 ": created " + 
				 localAnchor("/fta/"+sysname,
					     "SysOffice"));
	    return true;
	}
	catch (Exception e) {
	    _myserver.logMessage(LOG_ERROR,
				 sysname+
				 ": create SysOffice failed: exc=" + e);
	    //e.printStackTrace();
	    return false;
	}
    }

    /**
     * @roseuid 383EAE44E051
     */
    public boolean removeSystem(String sysname) {
	if (sysname != null) {
	    _offices.remove(sysname);
	    _myserver.logMessage(LOG_SUCCESS,
				 sysname+": remove SysOffice");
	    return true;
	}
	_myserver.logMessage(LOG_ERROR,
			     sysname+
			     ": remove SysOffice failed (nonexistant)");
	return false;
    }

    /**
     * @roseuid 38B3D2CFF862
     */
    private FTASysSub getSystem(String sysname)
	throws FTAUnknownObjectException {
	FTASysSub ss = (FTASysSub)_offices.get(sysname);
	if (ss==null) {
	    throw new FTAUnknownObjectException("No such system "+sysname);
	}
	return ss;
    }
  
    /**
     * this is sent by the first SysSub for system control
     * the available functions are constants in FTASysSub
     * @roseuid 38B3D2CFF864
     */
    /*package*/ boolean controlSystem(String sysname, int function,
				      int code) {
	try {
	    // wait until create has been done
	    _createsema.P();

	    FTASysSub ss = getSystem(sysname);
	    ss.controlSystem(function,code);
	    _myserver.logMessage(LOG_SUCCESS,
				 sysname+
				 ": control "+
				 localAnchor("/"+sysname+"/",
					     "System") + 
				 ": function=" + function +
				 ", code="+code);
	    _createsema.V(1);
	    return true;
	}
	catch (Exception e) {
	    _myserver.logMessage(LOG_ERROR,
				 sysname+
				 ": control system failed: exc="+e);
	    _createsema.V(1);
	    return false;
	}
    }
  
    private class SysThread extends Thread {
	private String _sysname;
	private String _mainhost;
	private String _acname;
	private String[] _arglist;
	private FTADaemon _daemon;
	private int _backport;
	private int _hostnum;
    
	public SysThread(FTADaemon daemon,String sysname,
			 String mainhost,
			 String acname,String[] arglist,
			 int backport,int hostnum) {
	    super(_mytg,sysname);
	    _daemon = daemon;
	    _sysname = sysname;
	    _mainhost = mainhost;
	    _acname = acname;
	    _arglist = arglist;
	    _hostnum = hostnum;
	    _backport = backport;
	}
    
	public void run() {
	    FTASysSub ss;
	    try {
		/* create new object
		 * this will throw an FTASysAccessException in case 
		 * of error the created system is supposed to register
		 * itself at the local registry, and initialize itself.
		 */
		ss = new FTASysSub(_daemon,_mainhost,_sysname,
				   _backport,_hostnum,
				    _acname,_arglist);
		// and put it into system list
		if (ss != null)
		    _offices.put(_sysname,ss);
	    }
	    catch (Exception e) {
		_createsema.V(1);
		//e.printStackTrace();
		_myserver.logMessage(LOG_ERROR,
				     _sysname+
				     ": creating System failed: exc="+e);
		return;
	    }

	    _createsema.V(1);
	    _myserver.logMessage(LOG_SUCCESS,
				 _sysname+": created " +
				 localAnchor("/fta/"+_sysname+"/",
					     "System") +
				 " done");

	    // wait for termination
	    int heartbeatrate = 
		_svc.getIntegerProperty("fta_heartbeatrate",10000);
	    try {
		for (;;) {
		    sleep(heartbeatrate);
		    int status = ss.getSystemStatus();
		    if (status == FTASysSub.FTACTLSHUTDOWN) {
			_myserver.logMessage(LOG_SUCCESS,
					     _sysname + 
					     ": shutdown");
			break;
		    }
		}
	    }
	    catch(InterruptedException e) {
		_myserver.logMessage(LOG_ERROR,
				     _sysname+": interrupted");
		// remove it from the systems list
		_offices.remove(_sysname);
	    }
      
	    // and terminate
	}
    }

    private class FTADaemonListener extends FTAListenerAdapter {

	public FTADaemonListener() {
	    super();
	    _ftadcmds.addListener(FTAMsg.FTAD_DAEMONREQ,this);
	    _ftadcmds.addListener(FTAMsg.FTAD_DAEMONREPLY,this);
	    _ftadcmds.addListener(FTAMsg.FTAD_SYSTEMREQ,this);
	    _ftadcmds.addListener(FTAMsg.FTAD_SYSTEMREPLY,this);
	    _ftadcmds.addListener(FTAMsg.FTAD_DIRREQ,this);
	    _ftadcmds.addListener(FTAMsg.FTAD_DIRREPLY,this);
	    _ftadcmds.addListener(FTAMsg.FTAD_SHUTDOWN,this);
	}

	public void notifyFtadShutdown() {
	    System.exit(0);
	}

	public void notifyFtadDaemonIdRequest(long time) {
	    _ftadcmds.sendMsg(FTADaemonReplyMsg.get(time,_myhostname));
	}

	public void notifyFtadDaemonIdReply(long time,String host) {
	    _daemons.remove(host);
	    _daemons.put(host,new Date(time));
	}

	public void notifyFtadSystemIdRequest(long time) {
	    Enumeration it = _offices.keys();
	    while (it.hasMoreElements()) {
		String name = (String)it.nextElement();
		String id = name + "@" + _myhostname;
		_ftadcmds.sendMsg(FTASystemReplyMsg.get(time,id));
	    }
	}

	public void notifyFtadSystemIdReply(long time,String sys) {
	    _remotesystems.remove(sys);
	    _remotesystems.put(sys,new Date(time));
	}

	public void notifyFtadDirIdRequest(long time) {
	    if (_isdirserver) {
		long t = System.currentTimeMillis();
		_ftadcmds.sendMsg(FTADirReplyMsg.get(t,_dirserver));
	    }
	}

	public void notifyFtadDirIdReply(long time,String url) {
	    if (_isdirserver) {
		// lose dirserver property
		if (url.equals(_dirserver))
		    return;
	    }

	    // set the dirserver url
	    _isdirserver = false;
	    _dirserver = url;
	}
    }

    /*************************************************************************/
    /* HTTP server output */
    /*************************************************************************/

    /**
     * HTML support for servlets
     */
    /*package*/ FTAServletReply replyText(String text) {
	FTAFile f = new FTAFile(text.length());
	f.add(text);
	return f;
    }

    /**
     * HTML support for servlets
     */
    /*package*/ FTAServletReply replyDone() {
	return replyText("DONE");
    }

    /**
     * HTML support for servlets
     */
    /*package*/ FTAServletReply replyFailed(String code) {
	return replyText("FAILED "+code);
    }


    /**
     * HTML support for servlets
     */
    public FTAServletReply redirect(String url) {
	FTAServletReply f = replyText("Redirected");
	f.redirect(url);
	return f;
    }

    /**
     * HTML support for servlets
     */
    public static String buildURL(String host,String path) {
	return "http://"+host+':'+_ftadport+path;
    }

    /**
     * HTML support for servlets
     */
    /*package*/ void addLink(FTAHtmlFile f,String path,String linktext) {
	f.anchor(path,linktext);
    }

    /**
     * HTML support for servlets
     */
    public static FTAServletReply stdFinish(FTAHtmlFile f) {
	return f.finish("http://ais.gmd.de/ARC/fta.html","FTA","/");
    }

    /**
     * HTML support for servlets
     */
    /*package*/ void addLocalLink(FTAHtmlFile f,String path,
				  String linktext) {
	f.anchor(_myurl+path,linktext);
    }

    /**
     * HTML support for servlets
     */
    /*package*/ void addItemizedLink(FTAHtmlFile f,String path,String linktext) {
	f.item();
	addLink(f,path,linktext);
    }

    /**
     * HTML support for servlets
     */
    /*package*/ void addItemizedLocalLink(FTAHtmlFile f,String path,
					    String linktext) {
	f.item();
	addLocalLink(f,path,linktext);
    }

    /*package*/ String anchor(String host,String path, String text) {
	return "<A HREF=\""+buildURL(host,path)+"\">"+text+"</A>";
    }

    /*package*/ String localAnchor(String path, String text) {
	return "<A HREF=\""+_myurl+path+"\">"+text+"</A>";
    }

    /**
     * HTML support for servlets
     */
    /*package*/ void addFormButton(FTAHtmlFile f,String button,
				   String cgi) {
	f.beginForm(_myurl+cgi);
	f.submitButton(button);
	f.endForm();
    }

    /**
     * create a local host key
     */
    /*package*/ void addLocalLinkList(FTAHtmlFile f,Hashtable tbl,
				      String linkname) {
	f.beginUL();
	Enumeration it = tbl.keys();
	boolean one = false;
	while (it.hasMoreElements()) {
	    String key = (String)it.nextElement();
	    Date date = (Date)tbl.get(key);
	    one = true;
	    addItemizedLocalLink(f,linkname+key+".html",key);
	    f.text(" ("+date+")\r\n");
	}
	if (one==false)
	    f.item("NONE");
	f.endUL();
    }

    /**
     * generate a remote host key
     */
    /*package*/ void addRemoteLinkList(FTAHtmlFile f,Hashtable tbl,
				       String prefix,String postfix) {
	f.beginUL();
	Enumeration it = tbl.keys();
	boolean one = false;
	while (it.hasMoreElements()) {
	    String key = (String)it.nextElement();
	    Date date = (Date)tbl.get(key);
	    one = true;
	    int atpos = key.indexOf('@');
	    String host;
	    String name;
	    if (atpos != -1) {
		host = key.substring(atpos+1);
		name = key.substring(0,atpos);
	    } else {
		host = key;
		name = "";
	    }
	    addItemizedLink(f,
			    buildURL(host,prefix + name + postfix),
			    key);
	    f.text(" ("+date+")\r\n");
	}
	if (one==false)
	    f.item("NONE");
	f.endUL();
    }

    /*package*/ boolean isDirServer() {
	return _isdirserver;
    }

    /*package*/ FTAFile listall(Hashtable ht) {
	FTAFile f = new FTAFile(100);
	Enumeration it = ht.keys();
	boolean first = true;
	while (it.hasMoreElements()) {
	    f.add((first?"":",") + (String)it.nextElement());
	    first = false;
	}
	f.add("\r\n");
	return f;
    }
    
    /**
     * get the maintained list of daemons in the network
     * @return hashtable
     */
    public Hashtable getDaemons() {
	return _daemons;
    }

    /**
     * get the maintained list of systems on this host
     * @return hashtable
     */
    /*package*/ Hashtable getLocalSystems() {
	return _offices;
    }

    /**
     * get the maintained list of offices on this host
     * @return hashtable
     */
    /*package*/ Hashtable getLocalOffices() {
	return _offices;
    }

    /**
     * get the maintained list of systems on the net
     * @return hashtable
     */
    /*package*/ Hashtable getRemoteSystems() {
	return _remotesystems;
    }

    /**
     * send message
     * @param msg message to be sent
     */
    /*package*/ void sendMsg(FTAMsg msg) {
	_ftadcmds.sendMsg(msg);
    }

    /**
     * get the directory server
     * @return dirserver URL
     */
    /*package*/ String getDirServer() {
	return _dirserver;
    }

    /**
     * get the latency constant
     * @return int
     */
    /*package*/ int getLatency() {
	return _latency;
    }

    /**
     *
     */
    public void logMessage(int type,String msg) {
	_myserver.logMessage(type,msg);
    }
}

/*
 * Local variables: 
 * tab-width: 8
 * c-basic-offset: 4
 * End: 
 */
