// Copyright (c)1999-2001 by "GMD-AiS"
//  Full name: "GMD National Research Center for Information Technology -
//  Institute for Autonomous Intelligent Systems", in future
//  named "Fraunhofer Institute for Autonomous Intelligent Systems"
//
// This software is provided 'as is' by the above copyright holder without
// any explicit or implicit warranty and support.
//
// Within the granted license, the licensee is allowed to freely use
// and modify the software EXCLUSIVELY FOR academical, SCIENTIFIC, and/or
// educational PURPOSES. It is permitted to cite (in the common scientific
// sense) fragments of the code in scientific publications of the licensee,
// and also use modified or unmodified parts of the software in own
// research prototypes.
//
// Commerical use, as well as integration into hardware/software systems or
// other products, other then for aforementioned scientific purposes, is
// PROHIBITED WITHOUT EXPLICIT PERMISSION of the copyright holder.
//
// Prototyping software that makes use of parts of this code have to
// refer to the above copyright when given out to third parties by
// a line
//   "This software contains or uses code that is copyrighted
//    by the GMD National Research Center for Information Technology
//    (Fraunhofer Institute for Autonomous Intelligent Systems)"
// in the user visible part of the software (documentation, about box, etc.)
//
// The copyright holder provides the software 'as is' without any support
// (request for bugfixes, for instance), and without any implicit or explicit
// warranty. In particular, the copyright holder cannot be made liable
// for lack of certain features, for damages of any kind that might result
// from any proper or improper use of this software. This software is not
// to be used in safety-critical environments.
//
// Beyond this license, GMD-AiS appreciates, in the scientific sense,
// reports and code submissions of significant changes or improvements
// of the jFTA core software that might be integrated into a future
// version of this code. Such reports are entirely voluntary, though;
// they are not mandated by this license. Use the address jfta@borneo.gmd.de
// (in future jfta@borneo.ais.fhg.de) for submissions.

// $Id: FTAContentHandler.java,v 1.3 2001/06/21 14:57:23 veit Exp $

package jfta.http;

import java.util.*;
import java.io.*;
import java.net.*;

import jfta.base.*;
import jfta.exception.*;

/**
 * The contenthandler class allows for additional control of the
 * way the output is processed. Its main purpose may be to 
 * tailor the output format to the formats the browser understands,
 * e.g. converting JPEG files to GIF, in case the browser does not
 * accept JPEGs. By default, the standard content handler just
 * leaves content-type and format untouched.
 */

public class FTAContentHandler {
    public final static int SC_OK = 200;
    public final static int SC_CREATED = 201;
    public final static int SC_ACCEPTED = 202;
    public final static int SC_NO_CONTENT = 204;
    public final static int SC_MOVED_PERMANENTLY = 301;
    public final static int SC_MOVED_TEMPORARILY = 302;
    public final static int SC_NOT_MODIFIED = 304;
    public final static int SC_BAD_REQUEST = 400;
    public final static int SC_UNAUTHORIZED = 401;
    public final static int SC_FORBIDDEN = 403;
    public final static int SC_NOT_FOUND = 404;
    public final static int SC_INTERNAL_SERVER_ERROR = 500;
    public final static int SC_NOT_IMPLEMENTED = 501;
    public final static int SC_BAD_GATEWAY = 502;
    public final static int SC_SERVICE_UNAVAILABLE = 504;

    protected String _contenttype;

    /**
     * mandatory default constructor
     */
    protected FTAContentHandler() {
    }

    /**
     * optional init method
     * @param server backlink to server
     * @param args argument vector
     */
    public void init(FTAHttpServer server,Vector args)
	throws FTAArgumentException {
	
	if (args != null && args.size() == 1) {
	    _contenttype = (String)args.get(0);
	}
    }
    
    /**
     * create a content handler with the specified type
     * @param contenttype specified MIME type
     */
    protected FTAContentHandler(String contenttype) {
	_contenttype = contenttype;
    }

    /**
     * return content type this handler manages
     * @return content type
     */
    public String getContentType() {
	return _contenttype;
    }

    /**
     * emit the HTML header to the browser.
     * @param the originating request
     * @param the reply provied by some servlet
     * @return true if the output of the body should be suppressed
     */
    public boolean emitHeader(FTAServletRequest req,
			      FTAServletReply reply)
	throws IOException {
	OutputStream out = req.getRawStream();
	String method = (String)req.getHeader().get("X-Method");
	boolean head = method.equals("HEAD");
	boolean options = method.equals("OPTIONS");

	int ecode = reply.getErrorCode();
	if (ecode == 0) ecode = 200;
	String emsg = getErrorMessage(ecode);
	int len = options ? 0 : reply.getContentLength();
	String contenttype = reply.getContentType();
	Date now = new Date();
	Hashtable header = reply.getCustomReply();
	FTAHttpServer server = req.getServer();

	StringBuffer sb = new StringBuffer();
	sb.append("HTTP/1.0 " + ecode + " " + emsg + "\r\n");
	sb.append("Date: " + now + "\r\n");
	sb.append("Server: " + server.getVersion() + "\r\n");
	sb.append("Connection: close\r\n");
	sb.append("Content-Length: " + len + "\r\n");
	sb.append("Content-Type : " + contenttype + "\r\n");
	if (options)
	    sb.append("Allow: HEAD, GET, POST, OPTIONS\r\n");
	if (header != null) {
	    Enumeration it = header.keys();
	    while (it.hasMoreElements()) {
		Object key = it.nextElement();
		String val = (String)header.get(key);
		sb.append(key + ": " + val + "\r\n");
	    }
	}
	sb.append("\r\n");
	String s = sb.toString();
	byte[] b = s.getBytes();
	out.write(b);
	return head || options;
    }
    
    /**
     * emit reply of the previously called servlet (in reply) to client
     * @param req request specifier
     * @param reply reply specifier
     */
    public void emit(FTAServletRequest req,FTAServletReply reply)
	throws IOException {
	InputStream in = reply.getStream();
	if (emitHeader(req,reply) == false) {
	    OutputStream out = req.getRawStream();
	    byte[] buf = new byte[1024];
	    int i;
	    while ((i=in.read(buf)) != -1)  {
		out.write(buf,0,i);
	    }
	}
	in.close();
    }

    /**
     * translate HTTP error code into string
     * @param ecode error code
     */
    public static String getErrorMessage(int ecode) {
	switch (ecode) {
	case SC_OK: return "Ok";
	case SC_CREATED: return "Created";
	case SC_ACCEPTED: return "Accepted";
	case SC_NO_CONTENT: return "No Content";
	case SC_MOVED_PERMANENTLY: return "Moved Permanently";
	case SC_MOVED_TEMPORARILY: return "Moved Temporarily";
	case SC_NOT_MODIFIED: return "Not Modified";
	case SC_BAD_REQUEST: return "Bad Request";
	case SC_UNAUTHORIZED: return "Unauthorized";
	case SC_FORBIDDEN: return "Forbidden";
	case SC_NOT_FOUND: return "Not Found";
	case SC_INTERNAL_SERVER_ERROR: return "Internal Server Error";
	case SC_NOT_IMPLEMENTED: return "Not Implemented";
	case SC_BAD_GATEWAY: return "Bad Gateway";
	case SC_SERVICE_UNAVAILABLE: return "Service Unavailable";
	default: return "Unknown Server Error "+ecode;
	}
    }

    /**
     * read contenthandlers from file
     * @param filename file to read definitions from
     * @return a hashtable
     */
    /**package*/ static Hashtable load(FTAHttpServer server,
				       String filename,Hashtable contents)
	throws IOException {

	if (contents == null) contents = new Hashtable();
	Properties p = new Properties();
	p.load(new FileInputStream(filename));

	Enumeration it = p.propertyNames();
	while (it.hasMoreElements()) {
	    String type = (String)it.nextElement();
	    String val = p.getProperty(type);

	    StringTokenizer st = new StringTokenizer(val," \t\n");
	    String clsname = st.nextToken();
	    Vector args = null;
	    if (st.hasMoreTokens()) {
		args = new Vector();
		while (st.hasMoreTokens()) {
		    args.add(st.nextToken());
		}
	    }

	    try {
		FTAContentHandler ch = byName(clsname);
		ch.init(server,args);
		contents.put(type,ch);
		//System.out.println("Contenthandler " + clsname + " loaded");
	    }
	    catch(FTAArgumentException e) {
		throw new IOException(clsname + 
		    ".init() failed: " + e);
	    }
	    catch(Exception e) {
		e.printStackTrace();
		throw new IOException("Creation of " + 
				      clsname + " failed: " + e);
	    }
	}
	return contents;
    }

    /**
     * Load a contenthandler by name
     * @param clsname java name of the handler, e.g. jfta.demo.mych
     * @return created ch instance
     * @exception Exception various exceptions while creating a new object 
     * may be raised
     */
    public static FTAContentHandler byName(String clsname)
	throws Exception {
	    Class chclass = Class.forName(clsname);
	    return (FTAContentHandler)chclass.newInstance();
    }

    /**
     * return the contenthandler classname
     * @return convert to string
     */
    public String toString() {
	return getClass().getName();
    }
}

/*
 * Local variables: 
 * tab-width: 8
 * c-basic-offset: 4
 * End: 
 */
