// Copyright (c)1999-2001 by "GMD-AiS"
//  Full name: "GMD National Research Center for Information Technology -
//  Institute for Autonomous Intelligent Systems", in future
//  named "Fraunhofer Institute for Autonomous Intelligent Systems"
//
// This software is provided 'as is' by the above copyright holder without
// any explicit or implicit warranty and support.
//
// Within the granted license, the licensee is allowed to freely use
// and modify the software EXCLUSIVELY FOR academical, SCIENTIFIC, and/or
// educational PURPOSES. It is permitted to cite (in the common scientific
// sense) fragments of the code in scientific publications of the licensee,
// and also use modified or unmodified parts of the software in own
// research prototypes.
//
// Commerical use, as well as integration into hardware/software systems or
// other products, other then for aforementioned scientific purposes, is
// PROHIBITED WITHOUT EXPLICIT PERMISSION of the copyright holder.
//
// Prototyping software that makes use of parts of this code have to
// refer to the above copyright when given out to third parties by
// a line
//   "This software contains or uses code that is copyrighted
//    by the GMD National Research Center for Information Technology
//    (Fraunhofer Institute for Autonomous Intelligent Systems)"
// in the user visible part of the software (documentation, about box, etc.)
//
// The copyright holder provides the software 'as is' without any support
// (request for bugfixes, for instance), and without any implicit or explicit
// warranty. In particular, the copyright holder cannot be made liable
// for lack of certain features, for damages of any kind that might result
// from any proper or improper use of this software. This software is not
// to be used in safety-critical environments.
//
// Beyond this license, GMD-AiS appreciates, in the scientific sense,
// reports and code submissions of significant changes or improvements
// of the jFTA core software that might be integrated into a future
// version of this code. Such reports are entirely voluntary, though;
// they are not mandated by this license. Use the address jfta@borneo.gmd.de
// (in future jfta@borneo.ais.fhg.de) for submissions.

//Source file: /home/rteams/jfta/jfta/hash/FTATagHash.java

package jfta.tagboard;

import jfta.base.*;
import jfta.exception.*;
import java.util.*;
import jfta.tags.*;

/**
 * The data structure to store tags is a hash of FTATagStores.
 * All relevant methods are synchronized.
 */
public class FTATagHash {
    /**
     * @serial the primary key structure
     */
    private Hashtable table;

    /**
     * @serial the current flip cycle
     */
    private long cycletime;
    /**
     * @serial reference to tagboard
     */
    private FTATagboardSub tagboard;

    /* helper class
     * This is the toplevel of a stored tag
     * which contains a reference to a tag and the number of tags
     * in that list. The tags are put in a linked list below this
     * structure. The sole purpose this link element is thus the
     * storage of the count, effectively, for use with getTagCount()
     * method.
     */ 
    private class FTATagLink extends Object {
	public FTATag link;
	public int count;

	public FTATagLink(FTATag tag) {
	    link = null;
	    add(tag);
	}

	public synchronized void add(FTATag tag) {
	    if (tag.link != null) {
		FTADebug.fatal("Tag relink error, tag="+tag);
		System.exit(1);
	    }
	    tag.link = link;
	    link = tag;
	    count++;
	}

	public synchronized void clear() {
	    FTATag it = link;
	    FTATag nx;
	    while (it != null) {
		nx = it.link;
		it.link = null;
		it = nx;
	    }
	    count = 0;
	}
    }
    
    /**
     * put a tag into hash
     * @param tag the tag to be stored
     */
    public synchronized void put(FTATag tag) {
	String tagtype = tag.getType();
	FTATagLink link = (FTATagLink)table.get(tagtype);
	if (link==null)
	    table.put(tagtype,new FTATagLink(tag));
	else
	    link.add(tag);
    }

    /**
     * provide an iterator over all tagstores
     * @return iterator
     */
    public Enumeration getStoreIterator() {
	return table.keys();
    }

    /**
     * create a new FTATagHash for the given cycletime
     * @param tb backlink to tagboard
     * @param ctime the current flip cycle
     * @roseuid 380B451BFC38
     */
    public FTATagHash(FTATagboardSub tb,long ctime) {
	table = new Hashtable();
	tagboard = tb;
	cycletime = ctime;
    }

    private class FTALinkIterator extends FTATagIterator {
	private FTATag nextObj;
	public FTALinkIterator(FTATag tlink,FTATagboardI tb) {
	    super(tb);
	    nextObj = tlink;
	}

	private boolean readNext() {
	    if (hasFlipped()) {
		nextObj=null;
		return false;
	    } else {
		nextObj = nextObj.link;
		return true;
	    }
	}

	public boolean hasNext() {
	    if (hasFlipped()) return false;
	    return nextObj != null;
	}
	
	public FTATag next2() {
	    FTATag tg = nextObj;
	    return readNext() ? tg : null;
	}
    }

    /**
     * get an iterator over all tags that match a certain pattern
     * @param tagtype tag type
     * @return a usable iterator
     * @roseuid 380EE17BB9C8
     */
    public synchronized FTATagIterator getIterator(String tagtype) {
	FTATagLink tlink = (FTATagLink)table.get(tagtype);
	return (tlink != null) ?
	    (FTATagIterator)new FTALinkIterator(tlink.link,tagboard) :
	    (FTATagIterator)new FTANullIterator(tagboard);
    }

    /**
     * get tag by hash
     * @param tagtype type of tag
     * @param hash tag hash value
     * @return tag or null
     */
    public FTATag getTagByHash(String tagtype,int hash) {
	FTATagLink tlink = (FTATagLink)table.get(tagtype);
	if (tlink != null) {
	    FTATag tg = tlink.link;
	    while (tg != null) {
		if (tg.hashCode()==hash) return tg;
		tg = tg.link;
	    }
	}
	return null;
    }

    /**
     * return the number of tags of a given type
     * @param tagtype the tagtype name
     * @return the number of patterns that match
     */
    public synchronized int getTagCount(String tagtype) {
	FTATagLink link = (FTATagLink)table.get(tagtype);
	return link != null ? link.count : 0;
    }

    /**
     * return all keys that are stored here
     */
    public synchronized String[] getTagTypes() {
	return (String[])table.keySet().toArray(new String[0]);
    }

    /**
     * clear tagboard
     */
    public synchronized void clear() {
	Enumeration it = table.elements();
	while (it.hasMoreElements()) {
	    FTATagLink link = (FTATagLink)it.nextElement();
	    link.clear();
	}
	table.clear();
    }

    /**
     * set the time
     */
    public synchronized void setTime(long ctime) {
	cycletime = ctime;
    }
}

/*
 * Local variables: 
 * tab-width: 8
 * c-basic-offset: 4
 * End: 
 */
